import abc
import datetime
import hashlib
import json
import random

from langchain_core.language_models import BaseChatModel


class Prompt:
    def __init__(self, prompt: str, ext_context: str = ''):
        self.prompt = Prompt.construct_prompt(prompt, ext_context)
        self.ext_context = ext_context

    @abc.abstractmethod
    def invoke(self, model: BaseChatModel): ...

    @staticmethod
    def construct_prompt(prompt: str, ext_context: str) -> str:
        __now = datetime.datetime.now().strftime('%m/%d/%Y %H:%M:%S.%f')
        __hash = hashlib.md5(f'{prompt}{__now}{random.uniform(0, 10 ** 3)}'.encode('utf-8')).hexdigest()
        __misc = {
            "__current_timestamp": __now,
            "__prompt_id": __hash,
        }
        prompt_json_part = prompt[prompt.find('{'):prompt.rfind('}') + 1]
        __prompt_dict = {
            '__misc': __misc,
            'prompt': json.loads(prompt_json_part)
        }
        prompt_addition = prompt[prompt.rfind('}') + 1:]
        if len(prompt_addition) > 0:
            __prompt_dict['additional_prompt'] = prompt_addition
        if len(ext_context) > 0:
            __prompt_dict['__ext_context'] = ext_context
        return json.dumps(__prompt_dict, ensure_ascii=False)



import abc
import hashlib
import inspect
import json
import logging
import random
import time
from typing import Callable

from langchain_core.language_models import BaseChatModel

from ceo.ability.agentic_ability import Ability, PREFIX as AGENTIC_ABILITY_PREFIX
from ceo.ability.base_ability import BaseAbility
from ceo.ability.mcp_ability import McpAbility
from ceo.message import AfterActionTakenMessage
from ceo.message.all_done_message import AllDoneMessage
from ceo.prompt import (
    SchedulerPrompt,
    RequestResolverPrompt,
    SelfIntroducePrompt
)

log = logging.getLogger('ceo')
SYSTEM_ABILITY_PREFIX = '__SystemAbility__'


class BaseAgent:
    def __init__(self, abilities: list[Callable], brain: BaseChatModel, name: str = '', request: str = ''):
        self._abilities = list()
        self._act_count = 0
        self._name = name
        self._model = brain
        self._request = self._request_by_step = str()
        if request is not None and request != '':
            self._request, self._request_by_step = RequestResolverPrompt(request).invoke(self._model)
        if self._name is None or len(self._name) < 1:
            self._name = self._generate_name()
        self.__prev_results = list()
        self.__schedule = list()
        self._introduction = str()
        self.grant_abilities(abilities)

    @property
    def abilities(self) -> list[Ability]:
        return self._abilities

    @property
    def name(self) -> str:
        return self._name

    @property
    def introduction(self) -> str:
        return self._introduction

    @property
    def brain(self) -> BaseChatModel:
        return self._model

    def __repr__(self):
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def __str__(self):
        return self.__repr__()

    def _generate_name(self) -> str:
        __abilities_ls = [ability.to_dict() for ability in self.abilities]
        __tmp_bytes = f'{__abilities_ls}{time.time()}{random.uniform(0, 10 ** 3)}'.encode('utf-8')
        __tmp_str = hashlib.md5(__tmp_bytes).hexdigest()
        __sample_str = ''.join(random.sample(__tmp_str, 6))
        return f'智能體{__sample_str}型號'

    def to_dict(self) -> dict:
        __model_dict = self._model.dict()
        model_name = __model_dict.get('model_name', __model_dict.get('_type', 'unknown'))
        return {
            "name": self.name,
            "brain": model_name,
            "abilities": [ability.to_dict() for ability in self.abilities]
        }

    def introduce(self, update: bool = True) -> str:
        # noinspection PyUnusedLocal
        def get_your_info(*args, **kwargs) -> dict:
            """
            What does this ability do: To get your personal information.
            In any case, the [__SystemAbility__get_your_info] ability can only be used once!!
            :return: All information of yourself.
            """
            _info_dict = self.to_dict()
            if 'brain' in _info_dict:
                del _info_dict['brain']
            return {
                'success': True,
                'info': _info_dict
            }

        get_your_info.__name__ = f'{SYSTEM_ABILITY_PREFIX}{get_your_info.__name__}'
        self.grant_ability(get_your_info, update_introduction=False)
        if update:
            self._introduction = SelfIntroducePrompt(agent=self).invoke(self._model)
        return self._introduction

    def grant_ability(self, ability: Callable | Ability, update_introduction: bool = True):
        abilities_except_mcp = [_a for _a in self._abilities if not isinstance(_a, McpAbility)]
        if isinstance(ability, Ability):
            for _ability in abilities_except_mcp:
                both_agentic = (_ability.name.startswith(AGENTIC_ABILITY_PREFIX)
                                and ability.name.startswith(AGENTIC_ABILITY_PREFIX))
                both_not_agentic = (not _ability.name.startswith(AGENTIC_ABILITY_PREFIX)
                                    and not ability.name.startswith(AGENTIC_ABILITY_PREFIX))
                if both_agentic:
                    if ability.name == _ability.name:
                        return
                elif both_not_agentic:
                    if inspect.getsource(ability.function) == inspect.getsource(_ability.function):
                        return
            self._abilities.append(ability)
        else:
            for _ability in abilities_except_mcp:
                if not _ability.name.startswith(AGENTIC_ABILITY_PREFIX):
                    if inspect.getsource(ability) == inspect.getsource(_ability.function):
                        return
            self._abilities.append(Ability(ability))
        self.introduce(update=update_introduction)

    def grant_abilities(self, abilities: list[Callable | Ability]):
        prev_size = len(self.abilities)
        for ability in abilities:
            self.grant_ability(ability, update_introduction=False)
        self.introduce(update=(prev_size != len(self.abilities)))

    def deprive_ability(self, ability: Callable | Ability, update_introduction: bool = True):
        abilities_except_mcp = [_a for _a in self._abilities if not isinstance(_a, McpAbility)]
        removed = False
        if not isinstance(ability, Ability):
            ability = Ability(ability)
        for _ability in abilities_except_mcp:
            both_agentic = (_ability.name.startswith(AGENTIC_ABILITY_PREFIX)
                            and ability.name.startswith(AGENTIC_ABILITY_PREFIX))
            both_not_agentic = (not _ability.name.startswith(AGENTIC_ABILITY_PREFIX)
                                and not ability.name.startswith(AGENTIC_ABILITY_PREFIX))
            if both_agentic:
                if _ability.name == ability.name:
                    self._abilities.remove(_ability)
                    removed = True
            elif both_not_agentic:
                if inspect.getsource(_ability.function) == inspect.getsource(ability.function):
                    self._abilities.remove(_ability)
                    removed = True
        self.introduce(update=(removed and update_introduction))

    def deprive_abilities(self, abilities: list[Callable | Ability]):
        prev_size = len(self.abilities)
        for ability in abilities:
            self.deprive_ability(ability, update_introduction=False)
        self.introduce(update=(prev_size != len(self.abilities)))

    def plan(self, _log: bool = True) -> list:
        scheduling = SchedulerPrompt(request=self._request_by_step, abilities=self._abilities)
        self.__schedule = scheduling.invoke(self._model)
        if _log:
            log.debug(f'Agent: {self._name}; Schedule: {[_.name for _ in self.__schedule]}; Request: "{self._request}";')
        return self.__schedule

    def reposition(self):
        self.__prev_results = list()
        self.__schedule = list()
        self._act_count = 0
        return self

    def assign(self, request: str):
        self._request, self._request_by_step = (
            RequestResolverPrompt(request=request).invoke(self._model))
        return self.reposition()

    def reassign(self, request: str):
        return self.assign(request)

    def relay(self, request: str, request_by_step: str):
        self._request = request
        self._request_by_step = request_by_step
        return self.reposition()

    @abc.abstractmethod
    def execute(self, args: dict, action: BaseAbility) -> AfterActionTakenMessage: ...

    @abc.abstractmethod
    def just_do_it(self, *args, **kwargs) -> AllDoneMessage: ...



from typing import Callable

from ceo.message.base_massage import BaseMessage
from ceo.brain.base_agent import BaseAgent
from ceo.util.synchronized_call import synchronized_call


class BaseHook(Callable[[BaseAgent, BaseMessage], BaseMessage]):

    def __init__(self, function: Callable[[BaseAgent, BaseMessage], BaseMessage]):
        self._function = function

    def __call__(self, agent: BaseAgent, message: BaseMessage):
        return synchronized_call(self._function, agent, message)

    @staticmethod
    def do_nothing() -> Callable:
        return lambda agent, msg: msg



import json
from dataclasses import dataclass

from typing_extensions import override

from ceo.message.base_massage import BaseMessage


@dataclass
class AllDoneMessage(BaseMessage):
    success: bool
    conclusion: str
    raw_response: str
    time_used: str | float | int
    step_count: int

    @override
    def to_dict(self):
        return {
            'success': self.success,
            'conclusion': self.conclusion,
            'raw_response': self.raw_response,
            'misc': {
                'time_used': self.time_used,
                'step_count': self.step_count
            }
        }

    @property
    def response_for_agent(self):
        _tmp = {
            'success': self.success,
            'response': self.raw_response
        }
        return json.dumps(_tmp, ensure_ascii=False)



import json
import logging

from langchain_core.language_models import BaseChatModel

from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class RequestResolverPrompt(Prompt):
    def __init__(self, request: str, ext_context: str = ''):
        prompt = json.dumps({
            "precondition": 'There is a user request shown below at <user_request>',
            "user_request": request,
            "task": "What you need to do is to tell user's intention based on <user_request>.",
            "additional_important": "For any details mentioned in <user_request>, you should preserve them in full, "
                                    "especially specific information with accuracy requirements "
                                    "such as all numbers occur, parameter, location, "
                                    "date/time, name, proper noun, entity, etc...",
            "hint_for_thinking": "Deduce and analyse the <user_request> step by step. "
                                 "Keep track of the steps' interdependence and orderliness.",
            "output_format": {
                "step_{n}": "({condition_for_step_{n}}) {action_of_step_{n}}"
            },
            "hint_1_for_output": "Break user's intention(s) down into possible steps. "
                                 "Keep track of the steps' interdependence and orderliness.",
            "hint_2_for_output": 'You must strictly follow the format in <output_format>! '
                                 'You can refer to example in <output_example>!',
            "limitation_1_for_output": 'Make your output brief, concise, accurate and short enough.',
            "limitation_2_for_output": 'Thought process is not required.',
            "output_example": {
                "step_1": "(Start) Open the door",
                "step_2": "(After: door opened) Go into the room",
                "step_3": "(After: walked in the room) Find the toys in the room",
                "step_...": "(After: found toys)..."
            }
        }, ensure_ascii=False)
        self.__request = request
        super().__init__(prompt, ext_context)
        log.debug(f'RequestResolverPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> tuple[str, str]:
        _dont_do_anything = "Don't do anything."
        if self.__request == '':
            return _dont_do_anything, _dont_do_anything
        user_request_by_step = model.invoke(self.prompt).content
        log.debug(f'RequestResolverResponse: {user_request_by_step}')
        return self.__request, user_request_by_step



from mcp import StdioServerParameters as StdioMcpConfig, ClientSession, stdio_client


def mcp_session(mcp_config: StdioMcpConfig | None):
    def decorator(func):
        async def wrapper(*args, **kwargs):
            if isinstance(mcp_config, StdioMcpConfig):
                async with stdio_client(mcp_config) as (read, write):
                    async with ClientSession(read, write) as session:
                        await session.initialize()
                        return await func(session, *args, **kwargs)
            else:
                ...
        return wrapper
    return decorator



import json
import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.exception.too_dumb_exception import TooDumbException
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')

END = '--END--'
SUCCESS = '--SUCCESS--'
FAILED = '--FAILED--'
THOUGHT_PROCESS = '--THOUGHT-PROCESS--'
CONCLUSION = '--CONCLUSION--'

OUTPUT_EXAMPLE = THOUGHT_PROCESS + """
(Start) [Grocery shopping]: Alex bought carrots and chicken breasts from the supermarket. """ + f'({SUCCESS})' + """
(After: Grocery shopping) [Cooking the meal]: I cooked the chicken with the carrots and garlic in a pan. """ + f'({SUCCESS})' + """
(After: Cooking the meal) [Plating the dish]: Sara plated the chicken and vegetables. """ + f'({SUCCESS})' + """
(After: Plating the dish) [Eating the meal]: There is no record in <history> indicating that someone sat down to eat the meal. """ + f'({FAILED})' + """
Based on above assessments, here is my conclusion:
""" + CONCLUSION + ("\nYour request has not been fully achieved. "
                    "I asked Alex to buy carrots and chicken breasts from the supermarket, "
                    "I cooked the chicken with the carrots and garlic in a pan, "
                    "Then I asked Sara to plate the chicken and vegetables, "
                    "but the meal was not eaten by anyone.\n") + END


class IntrospectionPrompt(Prompt):
    def __init__(self, request: str, history: dict | list, self_info: str | dict, ext_context: str = ''):
        prompt = json.dumps({
            "precondition": "Below in <history> are actions have been performed to achieve <request>. ",
            "request": request,
            "your_identity": self_info,
            "hint_for_your_identity": 'You can know your identity from <your_identity>.',
            "task": "Think step by step concisely whether <request> has been fully achieved "
                    "according to <history> and <request>. "
                    "Then, provide the detailed results mentioned in <history> accurately. "
                    "Finally, if the <request> has not been fully achieved, explain why failed?",
            "history": history,
            "output_datatype": "text",
            "output_format": f'{THOUGHT_PROCESS}\n'
                             '({condition_for_action_1}) [{action_1_to_take}]: {record_of_action_1_from_history} ({success_or_failed})\n'
                             '({condition_for_action_2}) [{action_2_to_take}]: {record_of_action_2_from_history} ({success_or_failed})\n'
                             '({condition_for_action_{n}}) [{action_{n}_to_take}]: {record_of_action_{n}_from_history} ({success_or_failed})\n'
                             'Based on above assessments, here is my conclusion:\n'
                             f'{CONCLUSION}\n'
                             '{conclusion}{results_of_actions}\n'
                             f'{END}',
            "output_example": OUTPUT_EXAMPLE,
            "limitation_1_for_output": 'You must strictly follow the format in <output_format>!! '
                                       'You should refer to example in <output_example>!!',
            "limitation_2_for_output": "Provide thought process (briefly and concisely) before conclusion.",
            "limitation_3_for_output": "Output should be concise, accurate, and short enough.",
            "limitation_for_conclusion": "In <conclusion> you must provide all the detailed results of all actions performed "
                                         "to achieve the <request>.",
            "hint_for_thought_process_pattern": f'The "{THOUGHT_PROCESS}" pattern marks the start of your thought process, '
                                                'do not forget to place it before your thought process.',
            "hint_for_thought_conclusion_pattern": f'The "{CONCLUSION}" pattern marks the start of your conclusion, '
                                                   'do not forget to place it before your conclusion.',
            "hint_for_end_pattern": f'The "{END}" pattern marks the end of your whole response, '
                                    f'no more words are allowed after "{END}" pattern. '
                                    f'The "{END}" pattern is absolutely important, do not forget to place it '
                                    'in the end of your response.'
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'IntrospectionPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel, stream: bool = False, max_retry: int = 6) -> tuple[str, str] | Iterator:
        if stream:
            return model.stream(self.prompt)
        count: int = 0
        exclamation = '!'
        tmp_prompt = self.prompt
        while True:
            # noinspection DuplicatedCode
            if count > 0:
                if count <= max_retry:
                    log.warning(f'IntrospectionPromptWarn: incorrectly formatted. Retry: {count}')
                else:
                    log.warning(f'IntrospectionPromptWarn: max retry exceeded.')
                    raise TooDumbException(model)
            count += 1
            resp = model.invoke(tmp_prompt).content
            log.debug(f"Introspection thought process: \n{resp}")
            if resp.count(THOUGHT_PROCESS) == 1 and resp.count(CONCLUSION) == 1 and resp.count(END) == 1:
                break
            tmp_prompt = (f'{self.prompt}Attention_{count}: '
                          f'You must strictly follow the format in <output_format>{count * 2 * exclamation} '
                          f'You should refer to example in <output_example>{count * 2 * exclamation}')
        log.debug(f'IntrospectionResponse: {resp}')
        _conclusion = resp[resp.rfind(CONCLUSION) + len(CONCLUSION):resp.rfind(END)].strip().strip('\n').strip('\r')
        return _conclusion, resp



import abc
import json

from typing_extensions import Callable


class BaseAbility(Callable):
    def __init__(self, name: str, description: str | dict, parameters: dict, returns: any):
        self._name = name
        self._description = description
        self._parameters = parameters
        self._returns = returns

    @abc.abstractmethod
    def __call__(self, *args, **kwargs): ...

    def __repr__(self):
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def __str__(self):
        return self.__repr__()

    def to_dict(self) -> dict:
        unnecessary_params: tuple = ('args', 'kwargs')
        return {
            'ability_name': self._name,
            'description': self._description,
            'parameters_required': [name for name in self._parameters.keys() if name not in unnecessary_params],
            'returns': str(self._returns)
        }

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description

    @property
    def parameters(self) -> dict:
        return self._parameters

    @property
    def returns(self) -> any:
        return self._returns



from .prompt import Prompt
from .analyser_prompt import AnalyserPrompt
from .scheduler_prompt import SchedulerPrompt
from .executor_prompt import ExecutorPrompt
from .introspection_prompt import IntrospectionPrompt
from .request_resolver_prompt import RequestResolverPrompt
from .self_introduce import SelfIntroducePrompt
from .docstring_prompt import DocstringPrompt
from .next_move_prompt import NextMovePrompt



import json
import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class SelfIntroducePrompt(Prompt):
    def __init__(self, agent: any, ext_context: str = ''):
        agent_info_dict = agent.to_dict()
        abilities_of_agent = agent_info_dict.get('abilities', [])
        for ability in abilities_of_agent:
            try:
                if isinstance(ability['description'], dict):
                    del ability['description']['parameters']
                    del ability['description']['returns']
                del ability['parameters_required']
                del ability['returns']
            except KeyError:
                pass
        agent_info_dict['abilities'] = abilities_of_agent
        prompt = json.dumps({
            "task": "Introduce yourself briefly based on the information provided. "
                    "Only tell what you exactly can do based on your abilities.",
            "hint": "Your name is shown in <your_name>",
            "your_name": agent_info_dict.get('name', None),
            "information": agent_info_dict,
            "output_format": "My name is <name>. What can I do: ..."
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'SelfIntroducePrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        if stream:
            return model.stream(self.prompt)
        resp = model.invoke(self.prompt).content
        log.debug(f'SelfIntroduceResponse: {resp}')
        return resp



from typing import Callable

from typing_extensions import override

from ceo.brain.hook.base_hook import BaseHook
from ceo.message.before_action_taken_message import BeforeActionTakenMessage


class BeforeActionTaken(BaseHook):
    def __init__(self, function: Callable):
        super().__init__(function)

    @override
    def __call__(self, agent: any, message: BeforeActionTakenMessage):
        return super().__call__(agent, message)



import json
import logging

from langchain_core.language_models import BaseChatModel

from ceo.ability.ability import Ability
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')

END = '--END--'


class SchedulerPrompt(Prompt):
    def __init__(self, request: str, abilities: list[Ability], ext_context: str = ''):
        self.abilities = abilities
        prompt = dict()
        for ability in self.abilities:
            ability_dict = ability.to_dict()
            if isinstance(ability_dict['description'], dict):
                try:
                    desc_dict = ability_dict['description']
                    if isinstance(desc_dict, str):
                        desc_dict = json.loads(desc_dict)
                    if isinstance(desc_dict, dict):
                        if 'parameters' in desc_dict.keys():
                            del desc_dict['parameters']
                        if 'returns' in desc_dict.keys():
                            del desc_dict['returns']
                        ability_dict['description'] = desc_dict
                except json.JSONDecodeError:
                    pass
            del ability_dict['parameters_required']
            prompt[ability.name] = ability_dict
        prompt = json.dumps({
            "precondition": "In <abilities> are the abilities you have. "
                            f'And there is a <user_request>.',
            "limitation": "You can only use abilities in <abilities>.",
            "user_request": request,
            "task": "What you need to do is to plan your workflow based on the <abilities> and <user_request>.",
            "description": "<user_request> might contains many steps, "
                           "think carefully about every step and plan your workflow "
                           "based on your abilities in <abilities>.",
            "hint_for_ability_choosing": "Some of the abilities may be irrelevant to <user_request>. "
                                         "Make sure to choose abilities properly and wisely.",
            "abilities": prompt,
            "output_format": "{thinking_process}\n"
                             "schedule:{schedule_as_a_list_of_ability_names}\n"
                             f"{END}",
            "hint_for_output": 'Firstly, output your thinking process step by step clear and organized. '
                               'Secondly, outputs a list of names of abilities, surrounded by "[ ]", split by ", ".',
            "limitation_for_thinking_process": 'Your thinking process should be brief, concise, and short enough.',
            "limitation_for_output": 'You must strictly follow the format in <output_format>!! '
                                     'You should refer to example in <output_example>!!',
            "output_example": "1.First, I need to determine which ingredients to purchase, "
                              "which requires checking a recipe or personal preferences to decide.\n"
                              "2.After determining the ingredients, "
                              "I need to go to the market or supermarket to buy the required ingredients.\n"
                              "3.After purchasing the ingredients, I need to bring them home.\n"
                              "4.Once home, I need to wash and prepare the ingredients.\n"
                              "5.After preparation, I start cooking.\n"
                              "6.After cooking is complete, I need to arrange the dishes on the dining table.\n"
                              "schedule:"
                              "[go_to_market, payment_purchase, go_home, do_wash, do_cook, arrange_dished_on_table]\n"
                              f"{END}",
            "limitation_for_end_pattern": f'The "{END}" pattern marks the end of your whole response, '
                                          f'no more words are allowed after "{END}" pattern. '
                                          f'The "{END}" pattern is absolutely important, do not forget to place it '
                                          'in the end of your response.'
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'SchedulerPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> list[Ability]:
        results = model.invoke(self.prompt).content
        log.debug(f'SchedulerResponse: {results}')
        results = results[results.rfind('['):results.rfind(']') + 1][1:-1].split(',')
        results = [result
                   .replace('\n', '')
                   .replace('\"', '')
                   .replace('\'', '').strip()
                   for result in results]
        _fin_results = list()
        for _a_result in results:
            for ability in self.abilities:
                if ability.name == _a_result:
                    _fin_results.append(ability)
        return _fin_results



import os

import langchain_openai
from langchain_core.language_models import BaseChatModel

DEFAULT_TMP = 0.125
DEFAULT_TOP_P = 1.00
DEFAULT_GPT = 'gpt-4o-mini'


def get_lm(key: str = None, model_name: str = DEFAULT_GPT, temp: float = DEFAULT_TMP, top_p: float = DEFAULT_TOP_P,
           base_url: str = None, stream: bool = False) -> BaseChatModel:
    if key is None:
        key = os.getenv('OPENAI_API_KEY', None)
    _kwargs = {
        'api_key': key,
        'model': model_name,
        'top_p': top_p,
        'temperature': temp,
        'streaming': stream
    }
    if base_url is not None:
        _kwargs['base_url'] = base_url
    return langchain_openai.ChatOpenAI(**_kwargs)



import abc
import copy
from collections import OrderedDict


class MemoryAugment:
    def __init__(self, memory: OrderedDict | None = None):
        self._memory: OrderedDict | None = None
        if memory is not None:
            self._memory = copy.deepcopy(memory)

    @property
    def memory(self) -> OrderedDict | None:
        if self._memory is not None:
            return copy.deepcopy(self._memory)
        return None

    @abc.abstractmethod
    def bring_in_memory(self, memory: OrderedDict): ...



from langchain_core.language_models import BaseChatModel


class TooDumbException(Exception):
    def __init__(self, model: BaseChatModel):
        __model_dict = model.dict()
        model_name = __model_dict.get('model_name', __model_dict.get('_type', 'unknown'))
        super().__init__(f'The model "{model_name}" is too dumb to be my brain, '
                         f'swap it or change its temperature.')



import json
import logging

from langchain_core.language_models import BaseChatModel

from ceo.ability.ability import Ability
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class AnalyserPrompt(Prompt):
    def __init__(self, request: str | dict, prev_results: list, action: Ability, ext_context: str = ''):
        self.action = action
        prev_results_str = str()
        if len(prev_results) > 0:
            prev_results_str += '\n'
        for result in prev_results:
            prev_results_str += f'{result};\n'
        prompt = {
            "precondition": "Below is the <tool(ability)> you can use (you can only use this tool(ability)). "
                            f'And there is a user request in <user_request>.',
            "user_request": request,
            "objective": "What you need to do is to generate values of parameters of the <tool(ability)> "
                         "to achieve <user request>",
            "hints": {
                "to_generate_parameters": "Parameters are described in <tool(ability)>, "
                                          "values of params are supposed to be generated by you, "
                                          "strictly abide by the parameter name and type.",
                "no_nonexistent_parameters": "You must not generate any nonexistent parameter, "
                                             "look carefully at parameters "
                                             "which are only described in <tool(ability)>.",
                "no_redundant_info": "With no any redundant information, only a json which provides parameters."
            },
            "tool(ability)": action.to_dict(),
            "output_format": "json",
            "output_example": '{"param_1.name": "value_for_param_1", "param_2.name": "value_for_param_2"}'
        }
        if len(prev_results_str) != 0:
            prompt['action_history (previous actions)'] = prev_results_str
            prompt['instruction_for_action_history'] = ("Generate response according to previous actions "
                                                        "performed by you.")
            prompt['hint_for_action_history'] = ("What <action_history> has shown is what you have done until now."
                                                 "What you need to do now is to think seriously "
                                                 "what you have not done yet "
                                                 "according to the <action_history> and <user request>."
                                                 "You are going to plan your next moves after <action_history>.")
        else:
            prompt['hint_for_first_move'] = 'This is your first step, you are starting to think.'
        prompt = json.dumps(prompt, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'AnalyserPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> tuple[Ability, dict]:
        # noinspection DuplicatedCode
        result = model.invoke(self.prompt).content
        log.debug(f'AnalyserResponse: {result}')
        if not result.startswith('{'):
            result = result[result.find('{'):]
        if not result.endswith('}'):
            result = result[:result.rfind('}') + 1]
        param = json.loads(result)
        return self.action, param



import inspect
import json
import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.ability.agentic_ability import PREFIX as AGENTIC_ABILITY_PREFIX
from ceo.ability.base_ability import BaseAbility
from ceo.message.after_action_taken_message import AfterActionTakenMessage
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class ExecutorPrompt(Prompt):
    def __init__(self, args: dict, action: BaseAbility, ext_context: str = ''):
        self.action = action
        self.args = dict()
        tmp_args = dict()
        for _k in args:
            self.args[_k] = args[_k]
            tmp_args[_k] = args[_k]
        if self.action.name.startswith(AGENTIC_ABILITY_PREFIX):
            tmp_args = {'arguments': 'Ask for a favor.'}
        prompt = json.dumps({
            "precondition": "Below is an ability shown at <ability> "
                            "and your arguments for using the <ability> is shown at <arguments>.",
            "task": "Explain what you are going to do.",
            "output_datatype": "text",
            "output_example": "I am trying to open calculator.",
            "ability": self.action.to_dict(),
            "arguments": tmp_args
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'ExecutorPrompt (before): {self.prompt}')

    def take_action(self):
        return self.action.__call__(**self.args)

    async def atake_action(self):
        return await self.action.__call__(**self.args)

    def explain(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        if stream:
            return model.stream(self.prompt)
        resp = model.invoke(self.prompt).content
        log.debug(f'ExecutorResponse (before): {resp}')
        return resp

    def summarize(self, result: any, model: BaseChatModel) -> AfterActionTakenMessage:
        tmp_args = self.args
        if self.action.name.startswith(AGENTIC_ABILITY_PREFIX):
            tmp_args = {'arguments': 'Ask for a favor.'}
        prompt = json.dumps({
            "precondition": "Below is an ability shown at <ability>, "
                            "your arguments for the <ability> is shown at <arguments>, "
                            "result of your using of this <ability> is shown at <result>.",
            "task": "Summarize what you have done according to <ability>, <arguments>, and <result> "
                    "accurately, comprehensively, and briefly.",
            "ability": self.action.to_dict(),
            "arguments": tmp_args,
            "result": str(result),
            "output_datatype": "text",
            "output_example": "I used the wechat_sender ability to wrote a wechat message which says 'Bonjour', "
                              "the result shows 'success' which indicates success of wechat message sending."
        }, ensure_ascii=False)
        if len(self.ext_context) > 0:
            prompt = Prompt.construct_prompt(prompt, self.ext_context)
        log.debug(f'ExecutorPrompt (after): {prompt}')
        return AfterActionTakenMessage(
            ability=self.action.name,
            arguments=tmp_args,
            returns=str(result),
            summarization=model.invoke(prompt).content
        )

    def invoke(self, model: BaseChatModel) -> AfterActionTakenMessage:
        return self.summarize(self.take_action(), model)

    async def ainvoke(self, model: BaseChatModel) -> AfterActionTakenMessage:
        if inspect.iscoroutinefunction(self.action.__call__):
            result = await self.atake_action()
            return self.summarize(result, model)
        return self.invoke(model)



import json
import logging
from collections import OrderedDict

from langchain_core.language_models import BaseChatModel

from ceo.ability import Ability
from ceo.ability.agentic_ability import PREFIX as AGENTIC_ABILITY_PREFIX
from ceo.message.before_action_taken_message import BeforeActionTakenMessage
from ceo.prompt.prompt import Prompt
from ceo.exception.too_dumb_exception import TooDumbException

log = logging.getLogger('ceo.prompt')

SEPARATOR = '--SEP--'
END = '--END--'
MISSION_COMPLETE = '-mission-complete-'
MISSION_FAILED = '-mission-failed-'
MAKE_FINAL_RESPONSE = '-make-final-response-'

OUTPUT_EXAMPLE = """
[step1] In the provided history, events related to the user's request are listed chronologically:
    1. Steve calculated the radius of the sphere using the expression "(3 * 174.9 / 15.9 * 2.77)", resulting in a radius of 91.41 cm.
    2. Steve then calculated the surface area of the sphere using the formula "4 * 3.14159 * (91.41^2)", resulting in a surface area of 105001.841348316 cm².
[step2] According to the history, the radius and surface area have been calculated, but the volume calculation and file writing have not been completed, therefore the user request has not been fully and properly accomplished.
[step3] The unfinished parts of the user request are the calculation of the volume and writing the results to 'result.txt'. The abilities I possess include "calculator" for performing calculations and "write_file" for writing content to a file. 
[step4] Since the user's request is not fully accomplished, and I have the ability to calculate the volume and write to a file, my next move is to use the "calculator" ability with the expression "(4/3) * 3.14159 * (91.41^3)" to compute the volume of the sphere.
[step5] This step is not applicable because the user request has not been fully accomplished, and I have the ability to continue progressing.
[step6] This step is not applicable because the user request has not been fully accomplished.
""" + SEPARATOR + """
args:{
  "{name_of_param}": "(4/3) * 3.14159 * (91.41^3)"
}
ability:[calculator]
""" + END


class NextMovePrompt(Prompt):
    def __init__(self, request: str | dict,
                 abilities: list[Ability],
                 history: OrderedDict | None = None,
                 ext_context: str = ''):
        self.abilities = abilities
        self.__ability_names = [MISSION_COMPLETE, MISSION_FAILED, MAKE_FINAL_RESPONSE]
        for ability in self.abilities:
            self.__ability_names.append(ability.name)
        abilities_dict: dict = {
            MISSION_COMPLETE: {
                'ability_name': MISSION_COMPLETE,
                'description': 'To be chosen only when mission is considered completed.',
                'parameters_required': [],
                'returns': None
            },
            MISSION_FAILED: {
                'ability_name': MISSION_FAILED,
                'description': 'To be chosen only when mission is considered failed.',
                'parameters_required': [],
                'returns': None
            },
            MAKE_FINAL_RESPONSE: {
                'ability_name': MAKE_FINAL_RESPONSE,
                'description': 'After ALL previous actions performed SUCCESSFULLY, '
                               'you can make the final response to user.',
                'parameters_required': [],
                'returns': None
            }
        }
        # noinspection PyUnusedLocal
        latest_progress = None
        for ability in self.abilities:
            abilities_dict[ability.name] = ability.to_dict()
        if history in ('', '[]', '()', '{}', {}, [], (), None) or len(history.keys()) == 0:
            latest_progress = history = "Nothing happened before you."
        else:
            latest_progress = history[list(history.keys())[-1]]
        prompt_dict = {
            "precondition": "In <abilities> are abilities you have, and there is a <user_request>. "
                            "<history> shows events happened before you, "
                            "<latest_progress> shows the latest progress of <user_request>.",
            "instructions_you_must_follow_step_by_step": [{
                    "step": 1,
                    "action": "List events from <history> and <latest_progress> "
                              "which could be related to <user_request> (respectively and chronologically).",
                    # "second_action": "Extract and list all information related to <user_request> from <history> "
                    #                  "formatted one by one respectively.",
                    "additional": "For all details mentioned in <history> about <user_request>, "
                                  "you should preserve them in full, "
                                  "especially specific information with accuracy requirements "
                                  "such as datas, numbers, dates, names, etc.",
                    "limitation_for_step1": "Make it brief, concise and accurate."
                }, {
                    "step": 2,
                    "action": "Analyse whether the <user_request> has been fully and properly accomplished "
                              "according to your report in <step_1>, "
                              "and provide your analysis process and basis briefly.",
                    "limitation_for_step2": "Make it brief, concise and accurate."
                }, {
                    "step": 3,
                    "condition": "If the <user_request> has not been fully properly accomplished",
                    "action": "Analyse whether your <abilities> can complete the unfinished part of the <user_request>, "
                              "and provide the basis briefly according to information provided in <step_1>."
                }, {
                    "step": 4,
                    "condition": "If the <user_request> has not been fully properly accomplished and "
                                 "there is an ability in your <abilities> "
                                 "that can further advance the accomplishment of the <user_request> based on <history>.",
                    "first_action": "Plan and explain your next move briefly based on <history> "
                                    "for further advancing the <user_request>.",
                    "second_action": "Choose and provide the ability according to your next move"
                                     "(only one single ability can be chosen)",
                    "third_action": "After you have chosen the ability as next move, "
                                    "generate arguments for the ability(function) to achieve <next move>, "
                                    "before you generate arguments, explain why you give these arguments briefly."
                }, {
                    "step": 5,
                    "condition": "If the <user_request> has not been fully properly accomplished and "
                                 "there is no ability in your <abilities> "
                                 "that can further advance the accomplishment of the <user_request>",
                    "action": f'Provide the special ability called "{MISSION_FAILED}" '
                              'with "args:{}".'
                }, {
                    "step": 6,
                    "condition": "If the <user_request> has been fully and properly accomplished "
                                 "according to <history> and <user_request>",
                    "action": f'Provide the special ability called "{MISSION_COMPLETE}" '
                              'with "args:{}".'
                }
            ],
            "be_aware_of_sequence_of_movements": 'Ensure that actions be taken in the proper order '
                                                 'according to <user_request>.',
            "output_format": "{step1_thought_process}\n{step2_thought_process}\n"
                             "{step3_thought_process}\n{step4_thought_process}\n"
                             "{step5_thought_process}\n{step6_thought_process}\n"
                             f"{SEPARATOR}\n"
                             'args:{'
                             '"{name_of_param_1}":{value_for_param_1},'
                             '"{name_of_param_2}":{value_for_param_2},'
                             '"{name_of_param_{n}}":{value_for_param_{n}}'
                             '}\n'
                             f'ability:[ability.name]\n{END}',
            "limitation_for_thought_process_output": "Thought processes of all steps(from 1 to 6) should be provided.",
            "limitation_for_args_output_format": f'The "{SEPARATOR}" pattern should be after '
                                                 f'all the thought processes and before the <args and ability>.'
                                                 f'The <args> should be after the "{SEPARATOR}" pattern.'
                                                 'The <args> should be well formatted in JSON!!'
                                                 'In <args> you only gives the arguments for the chosen one ability.'
                                                 'The <ability> should be after <args>.',
            "hint_for_separation_pattern": f'The "{SEPARATOR}" pattern which separates <thought processes> and '
                                           '<args and ability> is absolutely important, do not forget to place it.',
            "hint_for_end_pattern": f'The "{END}" pattern marks the end of your whole response, '
                                    f'no more words are allowed after "{END}" pattern. '
                                    f'The "{END}" pattern is absolutely important, do not forget to place it '
                                    'in the end of your response.',
            "output_example": OUTPUT_EXAMPLE,
            "hint_for_output": 'You must strictly follow the format in <output_format>! '
                               'You should refer to example in <output_example>!',
            "user_request": request,
            "history": history,
            "latest_progress": latest_progress,
            "hint_for_latest_progress": "The <latest_progress> shows the previous move.",
            "abilities": abilities_dict,
            "limitation_1_for_ability_choosing": "Only one ability can be chosen.",
            "limitation_2_for_ability_choosing": "You can only use the abilities listed in <abilities>.",
            "limitation_for_ability_output_format": 'The ability should be after the args. '
                                                    'The ability name should be surrounded by "[ ]".',
            "limitation_for_args": f'You must make sure the parameter_names you provide '
                                   f'for <args> are real and correct according to <abilities>!'
        }
        prompt = json.dumps(prompt_dict, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'NextMovePrompt: {self.prompt}')

    # noinspection PyUnusedLocal,DuplicatedCode
    def invoke(self, model: BaseChatModel, max_retry: int = 6) -> BeforeActionTakenMessage | bool:
        result: str = str()
        count: int = 0
        exclamation = '!'
        tmp_prompt = self.prompt
        while True:
            # noinspection DuplicatedCode
            if count > 0:
                if count <= max_retry:
                    log.warning(f'NextMovePromptWarn: incorrectly formatted. Retry: {count}')
                else:
                    log.warning(f'NextMovePromptWarn: max retry exceeded.')
                    raise TooDumbException(model)
            count += 1
            result = model.invoke(tmp_prompt).content
            log.debug(f"Next move thought process: \n{result}")
            _accurate_action_str = result[result.rfind(SEPARATOR) + len(SEPARATOR):result.rfind(END)]
            if (result.count(SEPARATOR) == 1
                    and result.count(END) == 1
                    and _accurate_action_str.count('ability:') == 1
                    and _accurate_action_str.count('args:') == 1):
                tmp_prompt_dict = None
                if isinstance(tmp_prompt, str):
                    tmp_prompt_dict = json.loads(tmp_prompt)
                tmp_prompt_dict_prompt_str = json.dumps(tmp_prompt_dict.get('prompt'), ensure_ascii=False)
                __additional_prompt = tmp_prompt_dict.get('additional_prompt', '')
                if len(__additional_prompt) > 0:
                    __additional_prompt += f"\n{'-' * 5}\n"
                tmp_prompt_dict_prompt_str += f"{__additional_prompt}"
                result = _accurate_action_str
                args = json.loads(result[result.find('{'):result.rfind('}') + 1].strip())
                result = result[result.rfind('}') + 1:]
                ability_name: str = result[result.find('['):result.rfind(']') + 1].strip()[1:-1]
                if ability_name not in self.__ability_names:
                    tmp_prompt = (f'{tmp_prompt_dict_prompt_str}There is no ability called "{ability_name}", '
                                  f'These abilities are available for you to choose: {self.__ability_names}.')
                    tmp_prompt = Prompt.construct_prompt(tmp_prompt, '')
                    continue
                if (ability_name.startswith(AGENTIC_ABILITY_PREFIX)
                        or MISSION_COMPLETE in ability_name
                        or MISSION_FAILED in ability_name
                        or MAKE_FINAL_RESPONSE in ability_name):
                    break
                _ability = None
                _wrong_param = False
                _wrong_param_names = list()
                for ability in self.abilities:
                    if ability.name == ability_name:
                        _ability = ability
                if _ability is not None:
                    for _k in args.keys():
                        if _k not in _ability.parameters.keys():
                            _wrong_param = True
                            _wrong_param_names.append(_k)
                if not _wrong_param:
                    break
                else:
                    tmp_prompt = (f'{tmp_prompt_dict_prompt_str}For ability called "{ability_name}", '
                                  f'these parameter_names are incorrect: {_wrong_param_names}, '
                                  f"correct parameter_names: {_ability.to_dict().get('parameters_required', [])};")
                    tmp_prompt = Prompt.construct_prompt(tmp_prompt, '')
                    continue
            tmp_prompt = (f'{self.prompt}Attention_{count}: '
                          f'You must strictly follow the format in <output_format>{count * 2 * exclamation} '
                          f'You should refer to example in <output_example>{count * 2 * exclamation}')
            tmp_prompt = Prompt.construct_prompt(tmp_prompt, '')
        if ability_name.__contains__(MISSION_COMPLETE) or ability_name.__contains__(MAKE_FINAL_RESPONSE):
            return True
        for ability in self.abilities:
            if ability.name == ability_name:
                return BeforeActionTakenMessage(ability=ability, arguments=args)
        return False



from typing import Callable


class NotARegularFunctionException(TypeError):
    def __init__(self, func: Callable):
        super().__init__(f'"{func.__qualname__}" is not a regular function. Abilities can only be regular functions.')



import hashlib
import json
import logging
import random
import datetime
import time
from typing import Callable
from typing_extensions import override
from collections import OrderedDict

from langchain_core.language_models import BaseChatModel

from ceo.ability.agentic_ability import PREFIX as AGENTIC_ABILITY_PREFIX
from ceo.ability.base_ability import BaseAbility
from ceo.brain.base_agent import BaseAgent
from ceo.brain.hook.before_action_taken import BeforeActionTaken
from ceo.brain.hook.after_action_taken import AfterActionTaken
from ceo.brain.hook.base_hook import BaseHook
from ceo.brain.memory_augment import MemoryAugment
from ceo.enum.Personality import Personality
from ceo.message.all_done_message import AllDoneMessage
from ceo.message.before_action_taken_message import BeforeActionTakenMessage
from ceo.message.after_action_taken_message import AfterActionTakenMessage
from ceo.prompt import (
    NextMovePrompt,
    ExecutorPrompt,
    IntrospectionPrompt
)

PRUDENT_P = 0.25
PRUDENT_BETA = 1.45
INQUISITIVE_P = 0.05
INQUISITIVE_BETA = 1.25
log = logging.getLogger('ceo')


class Agent(BaseAgent, MemoryAugment):
    def __init__(self, abilities: list[Callable],
                 brain: BaseChatModel, name: str = '',
                 personality: Personality = Personality.PRUDENT,
                 request: str = '', memory: OrderedDict | None = None):
        BaseAgent.__init__(self, abilities=abilities, brain=brain, name=name, request=request)
        MemoryAugment.__init__(self, memory=memory)
        self._expected_step = 0
        if personality == Personality.PRUDENT:
            self._p = self.__base_p = PRUDENT_P
            self._beta = PRUDENT_BETA
        elif personality == Personality.INQUISITIVE:
            self._p = self.__base_p = INQUISITIVE_P
            self._beta = INQUISITIVE_BETA

    @property
    def p(self) -> float:
        return self._p

    @property
    def base_p(self) -> float:
        return self.__base_p

    @property
    def beta(self) -> float:
        return self._beta

    @override
    def bring_in_memory(self, memory: OrderedDict):
        self._memory.update(memory)
        log.debug(f'Agent: {self._name}; '
                  f'Memory brought in: {len(memory.keys())};')
        return self

    @override
    def reposition(self):
        BaseAgent.reposition(self)
        self._memory = OrderedDict()
        self._expected_step = 0
        self._p = self.__base_p
        return self

    @override
    def assign(self, request: str):
        BaseAgent.assign(self, request)
        return self.reposition()

    @override
    def reassign(self, request: str):
        return self.assign(request)

    @override
    def relay(self, request: str, request_by_step: str):
        self._request = request
        self._request_by_step = request_by_step
        return self.reposition()

    @override
    def execute(self, args: dict, action: BaseAbility) -> AfterActionTakenMessage:
        return ExecutorPrompt(action=action, args=args).invoke(self.brain)

    @override
    def just_do_it(self, *args, **kwargs) -> AllDoneMessage:
        __after_action_taken_hook: AfterActionTaken | Callable = BaseHook.do_nothing()
        __before_action_taken_hook: BeforeActionTaken | Callable = BaseHook.do_nothing()
        for _arg in args:
            if isinstance(_arg, AfterActionTaken):
                __after_action_taken_hook = _arg
            if isinstance(_arg, BeforeActionTaken):
                __before_action_taken_hook = _arg
        __start_time = time.perf_counter()
        if self._expected_step < 1:
            self.estimate_step()
        log.debug(f'Agent: {self._name}; Expected steps: {self._expected_step}; Request: "{self._request}";')
        stop = False
        while True:
            if self._act_count > self._expected_step:
                stop = self.stop()
                self.penalize()
            next_move = False
            if not stop:
                combined_request = {
                    'raw_request': self._request,
                    'request_by_step': self._request_by_step
                }
                next_move = NextMovePrompt(
                    request=combined_request,
                    abilities=self._abilities,
                    history=self.memory
                ).invoke(self._model)
                if isinstance(next_move, BeforeActionTakenMessage):
                    next_move = __before_action_taken_hook(self, next_move)
                    action, args = next_move.ability, next_move.arguments
                    if action.name.startswith(AGENTIC_ABILITY_PREFIX):
                        args = {
                            'request': self._request,
                            'request_by_step': self._request_by_step,
                            'memory': self.memory,
                            'before_action_taken_hook': __before_action_taken_hook,
                            'after_action_taken_hook': __after_action_taken_hook
                        }
                    __after_execution_msg = self.execute(args=args, action=action)
                    self.memorize(__after_action_taken_hook(self, __after_execution_msg))
                    self._act_count += 1
                    continue
            brief_conclusion, response = IntrospectionPrompt(
                request=self._request,
                history=self.memory,
                self_info=self.introduction
            ).invoke(self._model)
            __time_used = time.perf_counter() - __start_time
            __step_count = self._act_count
            self.reposition()
            log.debug(f'Agent: {self._name}; Conclusion: {brief_conclusion};')
            log.debug(f'Agent: {self._name}; Step count: {__step_count}; Time used: {__time_used} seconds;')
            return AllDoneMessage(
                success=next_move,
                conclusion=brief_conclusion,
                raw_response=response,
                time_used=__time_used,
                step_count=__step_count
            )

    def assign_with_memory(self, request: str, memory: OrderedDict):
        return self.assign(request).bring_in_memory(memory)

    def estimate_step(self):
        if self._request_by_step == '':
            self._expected_step = 0
            return
        self._expected_step = len(self.plan(_log=False))
        return self

    def set_expected_step(self, expected_step: int):
        self._expected_step = expected_step
        return self

    def memorize(self, action_taken: AfterActionTakenMessage):
        if action_taken is None:
            return self
        now = datetime.datetime.now().strftime('%m/%d/%Y %H:%M:%S.%f')
        _tmp_action_taken = action_taken.to_dict()
        if 'summarization' in _tmp_action_taken.keys():
            del _tmp_action_taken['summarization']
        if _tmp_action_taken.get('ability', str()).startswith(AGENTIC_ABILITY_PREFIX):
            if 'arguments' in _tmp_action_taken.keys():
                _tmp_action_taken['arguments'] = 'Ask for a favor.'
        new_memory = {
            "timestamp": now,
            "agent_name": self._name,
            f"message_from_{self._name}": action_taken.summarization,
            f'action_taken_by_{self._name}': _tmp_action_taken
        }
        mem_hash = hashlib.md5(json.dumps(new_memory, ensure_ascii=False).encode()).hexdigest()
        self._memory[f"agent:[{self._name}] at:[{now}] hash:[{mem_hash}]"] = new_memory
        log.debug(f'Agent: {self._name}; Memory size: {len(self._memory.keys())}; Memory update: {action_taken.summarization};')
        return self

    def stop(self) -> bool:
        resample = 3
        log.debug(f'Agent: {self._name}; Termination Probability(p): {self._p}; Penalty Rate(beta): {self._beta};')
        rand_sum = 0.0
        for i in range(resample):
            rand_sum += random.uniform(0, 1)
        rand_avg = rand_sum / resample
        if rand_avg > self._p:
            return False
        return True

    def penalize(self):
        self._p = (self._beta * self._p) % 1.0
        return self

    def set_penalty(self, p: float, beta: float):
        self._p = self.__base_p = p
        self._beta = beta
        return self

    def change_personality(self, personality: Personality):
        if personality == Personality.PRUDENT:
            return self.set_penalty(p=PRUDENT_P, beta=PRUDENT_BETA)
        elif personality == Personality.INQUISITIVE:
            return self.set_penalty(p=INQUISITIVE_P, beta=INQUISITIVE_BETA)
        return self



import inspect
import json
import logging
from collections.abc import Iterator
from typing import Callable

from langchain_core.language_models import BaseChatModel

from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class DocstringPrompt(Prompt):
    def __init__(self, function: Callable, ext_context: str = ''):
        signature = inspect.signature(function)
        f_name: str = function.__name__
        f_parameters: dict = dict()
        f_returns: str = str(signature.return_annotation)
        for name, param in signature.parameters.items():
            f_parameters[name] = str(param.annotation)
        function_repr = {
            'name': f_name,
            'parameters': f_parameters,
            'returns': f_returns,
            'source_code': inspect.getsource(function)
        }
        docstring_format = {
            "description": {
                "brief_description": "{Brief and accurate description of the function, "
                                     "including its input and output, main logic, algorithm used, "
                                     "precautions mentioned, and any other relevant information.}",
                "parameters": [{
                        "{name of param_1}": {
                            "name": "{name of param_1}",
                            "type": "{data type of param_1}",
                            "description": "{Brief and accurate description of param_1, "
                                           "including its role in the function "
                                           "and the constraints for it.}"
                        }
                    }, {
                        "{name of param_2}": {
                            "name": "{name of param_2}",
                            "type": "{data type of param_2}",
                            "description": "{Brief and accurate description of param_2, "
                                           "including its role in the function "
                                           "and the constraints for it.}"
                        }
                    }, {
                        "{name of param_...}": {
                            "name": "{name of param_...}",
                            "type": "{data type of param_...}",
                            "description": "{Brief and accurate description of param_..., "
                                           "including its role in the function "
                                           "and the constraints for it.}"
                        }
                    }
                ],
                "returns": {
                    "type": "{data type of the return value}",
                    "description": "{Brief and accurate description of the return value, "
                                   "including its meaning.}"
                }
            }
        }
        prompt = json.dumps({
            "objective": "Generate description for the <function>.",
            "function": function_repr,
            "hint_for_function_understanding": "The <source_code> provides the unabridged definition of the <function>."
                                               "You can fully understand the <function> by reading <source_code>.",
            "output_format": "json",
            "hint_for_output_format": docstring_format
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'DocstringPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> str | Iterator:
        # noinspection DuplicatedCode
        raw_docstring = model.invoke(self.prompt).content
        log.debug(f'DocstringResponse: {raw_docstring}')
        if not raw_docstring.startswith('{'):
            raw_docstring = raw_docstring[raw_docstring.find('{'):]
        if not raw_docstring.endswith('}'):
            raw_docstring = raw_docstring[:raw_docstring.rfind('}') + 1]
        return raw_docstring.replace('\n', '')



from dataclasses import dataclass

from typing_extensions import override

from ceo.ability.ability import Ability
from ceo.message.base_massage import BaseMessage


@dataclass
class BeforeActionTakenMessage(BaseMessage):
    ability: Ability
    arguments: dict

    @override
    def to_dict(self):
        return {
            'ability': self.ability.to_dict(),
            'args': self.arguments
        }



import asyncio
import inspect
import threading

from typing_extensions import Callable


def synchronized_call(func: Callable, *args, **kwargs):
    if inspect.iscoroutinefunction(func):
        __res = None

        def __func(loop: asyncio.AbstractEventLoop):
            nonlocal __res, args, kwargs
            try:
                __res = loop.run_until_complete(func(*args, **kwargs))
            finally:
                loop.close()

        __thread = threading.Thread(
            target=__func,
            args=(asyncio.new_event_loop(),)
        )
        __thread.start()
        __thread.join(timeout=None)
        return __res
    return func(*args, **kwargs)


def sync_call(func: Callable):
    def wrapper(*args, **kwargs):
        return synchronized_call(func, *args, **kwargs)
    return wrapper



import inspect
import json

from typing_extensions import Callable, override

from ceo.util.synchronized_call import synchronized_call
from ceo.ability.base_ability import BaseAbility


class Ability(BaseAbility):
    def __init__(self, function: Callable):
        signature = inspect.signature(function)
        doc_str = inspect.getdoc(function)
        if doc_str is None:
            doc_str = json.dumps({
                'src': inspect.getsource(function)
            }, ensure_ascii=False)
        _description: str | dict = str()
        self._function: Callable = function
        _parameters: dict = dict()
        for name, param in signature.parameters.items():
            _parameters[name] = str(param.annotation)
        try:
            _description = json.loads(doc_str)
            _description = _description.get('description', _description)
        except json.decoder.JSONDecodeError:
            _description = doc_str
        except AttributeError:
            _description = doc_str
        super().__init__(
            name=function.__name__,
            description=_description,
            parameters=_parameters,
            returns=signature.return_annotation
        )

    @override
    def __call__(self, *args, **kwargs):
        return synchronized_call(self.function, *args, **kwargs)

    @property
    def function(self) -> Callable:
        return self._function



import os

import langchain_community.chat_models.tongyi
from langchain_core.language_models import BaseChatModel

DEFAULT_TMP = 0.125
DEFAULT_TOP_P = 1.00
DEFAULT_QWEN = 'qwen-plus'


def get_lm(key: str = None, model_name: str = DEFAULT_QWEN, temp: float = DEFAULT_TMP, top_p: float = DEFAULT_TOP_P, stream: bool = False) -> BaseChatModel:
    if key is None:
        key = os.getenv('DASHSCOPE_API_KEY', os.getenv('OPENAI_API_KEY', None))
    _kwargs = {
        'dashscope_api_key': key,
        'model': model_name,
        'top_p': top_p,
        'temperature': temp,
        'streaming': stream
    }
    return langchain_community.chat_models.tongyi.ChatTongyi(**_kwargs)



import logging
import time
from collections import OrderedDict

from typing_extensions import Callable, override
from langchain_core.language_models import BaseChatModel
from mcp import ClientSession

from ceo.ability.agentic_ability import PREFIX as AGENTIC_ABILITY_PREFIX
from ceo.ability.agentic_ability import Ability
from ceo.ability.base_ability import BaseAbility
from ceo.ability.mcp_ability import McpAbility
from ceo.brain.agent import Agent
from ceo.brain.hook import AfterActionTaken, BeforeActionTaken
from ceo.brain.hook.base_hook import BaseHook
from ceo.enum import Personality
from ceo.message import AllDoneMessage, AfterActionTakenMessage, BeforeActionTakenMessage
from ceo.prompt import ExecutorPrompt, IntrospectionPrompt, NextMovePrompt

log = logging.getLogger('ceo')


class McpAgent(Agent):
    def __init__(self, session: ClientSession,
                 brain: BaseChatModel, name: str = '',
                 personality: Personality = Personality.PRUDENT,
                 request: str = '', memory: OrderedDict | None = None):
        super().__init__(abilities=[], brain=brain,
                         name=name, personality=personality,
                         request=request, memory=memory)
        self._session = session

    async def fetch_abilities(self):
        async def __fetch() -> list[McpAbility]:
            _tmp_tools = (await self.session.list_tools()).tools
            mcp_tools = [McpAbility(mcp_tool=_tool, session=self.session) for _tool in _tmp_tools]
            return mcp_tools

        self.grant_abilities(await __fetch())
        return self

    @override
    def grant_ability(self, ability: Callable | Ability | McpAbility, update_introduction: bool = True):
        if isinstance(ability, McpAbility):
            self._abilities.append(ability)
            self.introduce(update=update_introduction)
        else:
            super().grant_ability(ability, update_introduction=update_introduction)

    @override
    def grant_abilities(self, abilities: list[Callable | Ability | McpAbility]):
        super().grant_abilities(abilities=abilities)

    @override
    def deprive_ability(self, ability: Callable | Ability | McpAbility, update_introduction: bool = True):
        removed = False
        if isinstance(ability, McpAbility):
            if ability in self.abilities:
                self._abilities.remove(ability)
                removed = True
            self.introduce(update=(removed and update_introduction))
        else:
            super().deprive_ability(ability, update_introduction=update_introduction)

    @override
    def deprive_abilities(self, abilities: list[Callable | Ability | McpAbility]):
        super().deprive_abilities(abilities=abilities)

    @override
    async def execute(self, args: dict, action: BaseAbility) -> AfterActionTakenMessage:
        return await ExecutorPrompt(action=action, args=args).ainvoke(self.brain)

    @override
    async def just_do_it(self, *args, **kwargs) -> AllDoneMessage:
        __after_action_taken_hook: AfterActionTaken | Callable = BaseHook.do_nothing()
        __before_action_taken_hook: BeforeActionTaken | Callable = BaseHook.do_nothing()
        for _arg in args:
            if isinstance(_arg, AfterActionTaken):
                __after_action_taken_hook = _arg
            if isinstance(_arg, BeforeActionTaken):
                __before_action_taken_hook = _arg
        __start_time = time.perf_counter()
        if self._expected_step < 1:
            self.estimate_step()
        log.debug(f'Agent: {self._name}; Expected steps: {self._expected_step}; Request: "{self._request}";')
        stop = False
        while True:
            if self._act_count > self._expected_step:
                stop = self.stop()
                self.penalize()
            next_move = False
            if not stop:
                combined_request = {
                    'raw_request': self._request,
                    'request_by_step': self._request_by_step
                }
                next_move = NextMovePrompt(
                    request=combined_request,
                    abilities=self._abilities,
                    history=self.memory
                ).invoke(self._model)
                if isinstance(next_move, BeforeActionTakenMessage):
                    next_move = __before_action_taken_hook(self, next_move)
                    action, args = next_move.ability, next_move.arguments
                    if action.name.startswith(AGENTIC_ABILITY_PREFIX):
                        args = {
                            'request': self._request,
                            'request_by_step': self._request_by_step,
                            'memory': self.memory,
                            'before_action_taken_hook': __before_action_taken_hook,
                            'after_action_taken_hook': __after_action_taken_hook
                        }
                    __after_execution_msg = await self.execute(args=args, action=action)
                    self.memorize(__after_action_taken_hook(self, __after_execution_msg))
                    self._act_count += 1
                    continue
            brief_conclusion, response = IntrospectionPrompt(
                request=self._request,
                history=self.memory,
                self_info=self.introduction
            ).invoke(self._model)
            __time_used = time.perf_counter() - __start_time
            __step_count = self._act_count
            self.reposition()
            log.debug(f'Agent: {self._name}; Conclusion: {brief_conclusion};')
            log.debug(f'Agent: {self._name}; Step count: {__step_count}; Time used: {__time_used} seconds;')
            return AllDoneMessage(
                success=next_move,
                conclusion=brief_conclusion,
                raw_response=response,
                time_used=__time_used,
                step_count=__step_count
            )

    @property
    def session(self) -> ClientSession:
        return self._session



from dataclasses import dataclass

from typing_extensions import override

from ceo.message.base_massage import BaseMessage


@dataclass
class AfterActionTakenMessage(BaseMessage):
    ability: str
    arguments: str | dict
    returns: str
    summarization: str

    @override
    def to_dict(self):
        return {
            'ability': self.ability,
            'arguments': self.arguments,
            'returns': self.returns,
            'summarization': self.summarization,
        }



import logging

# noinspection PyUnresolvedReferences
from .util.mcp_session import StdioMcpConfig
from .brain.lm import (
    get_openai_model,
    get_dashscope_model,
    get_deepseek_model
)
from .brain.agent import Agent
from .brain.mcp_agent import McpAgent
from .ability import Ability, McpAbility, AgenticAbility
from .util.agentic import agentic
from .util.ability import ability
from .util.synchronized_call import synchronized_call, sync_call
from .util.mcp_session import mcp_session
from .enum import Personality

__AUTHOR__ = '吴子豪 / Vortez Wohl'
__EMAIL__ = 'vortez.wohl@gmail.com'
__VERSION__ = '0.13.0-preview'
__GITHUB__ = 'https://github.com/vortezwohl'
__BLOG__ = 'https://vortezwohl.github.io'

logger = logging.getLogger('ceo')
logger.setLevel(logging.ERROR)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('[%(levelname)s] %(asctime)s %(name)s : %(message)s')
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)

logger = logging.getLogger('ceo.prompt')
logger.setLevel(logging.ERROR)
logger = logging.getLogger('ceo.ability')
logger.setLevel(logging.ERROR)
logger = logging.getLogger('ceo.agent')
logger.setLevel(logging.ERROR)



import inspect
import json
import logging
import os
from typing import Callable

from langchain_core.language_models import BaseChatModel

from ceo.brain.lm import get_openai_model
from ceo.prompt import DocstringPrompt
from ceo.exception.not_a_regular_function_exception import NotARegularFunctionException

log = logging.getLogger('ceo.ability')


def ability(brain: BaseChatModel, cache: bool = True, cache_dir: str = ''):
    # noinspection PyShadowingNames
    def docstring_generator(func: Callable, brain: BaseChatModel, cache: bool, cache_dir: str) -> Callable:
        func.__doc__ = DocstringPrompt(func).invoke(brain)
        log.debug(f'Docstring generated for {func.__name__}. Docstring: "{func.__doc__}"')
        if cache:
            cache_function(func, cache_dir)
        return func

    # noinspection PyShadowingNames
    def make_cache_filename(func: Callable, cache_dir: str, create_path: bool = False) -> str:
        func_source_file_long = inspect.getfile(func)
        func_source_file_short = os.path.basename(func_source_file_long)
        func_source_file_short = func_source_file_short[:func_source_file_short.rfind('.py')]
        func_source_path = os.path.dirname(func_source_file_long)
        cache_path = os.path.join(func_source_path, cache_dir)
        if create_path and not os.path.exists(cache_path):
            os.mkdir(cache_path)
        return os.path.join(cache_path, f'{func_source_file_short}.{func.__name__}.cache')

    def get_source(func: Callable) -> str:
        source_lines = inspect.getsourcelines(func)[0]
        for i, line in enumerate(source_lines):
            if f'@{ability.__name__}' in line:
                del source_lines[i]
                break
        return str().join(source_lines)

    # noinspection PyShadowingNames
    def cache_function(func: Callable, cache_dir: str) -> Callable:
        cache_file = make_cache_filename(func, cache_dir, create_path=True)
        func_doc = inspect.getdoc(func)
        try:
            func_doc = json.loads(func_doc)
        except json.decoder.JSONDecodeError:
            pass
        cache_data: dict = {
            'src': get_source(func),
            'doc': func_doc
        }
        with open(cache_file, 'wb') as f:
            cache_data_str = json.dumps(cache_data, ensure_ascii=False)
            cache_data_bytes = cache_data_str.encode('utf-8')
            f.write(cache_data_bytes)
        return func

    # noinspection PyShadowingNames
    def read_cache(func: Callable, cache_dir: str) -> dict:
        cache_file = make_cache_filename(func, cache_dir)
        cache_data: dict = dict()
        if not os.path.exists(cache_file):
            return cache_data
        with open(cache_file, 'rb') as f:
            cache_data_bytes = f.read()
            cache_data = json.loads(cache_data_bytes)
        return cache_data

    def check_if_function_is_method(func: Callable) -> Callable:
        if '.' in func.__qualname__:
            raise NotARegularFunctionException(func)
        return func

    if cache_dir in ('', None):
        cache_dir = '.cache'

    if callable(brain) and not isinstance(brain, BaseChatModel):
        # noinspection DuplicatedCode
        def decorator(func):
            func = check_if_function_is_method(func)
            cache_func = read_cache(func, cache_dir)
            if cache and cache_func.get('src', None) == get_source(func):
                func.__doc__ = json.dumps(cache_func.get('doc'), ensure_ascii=False)
                return func
            return docstring_generator(func, get_openai_model(), cache, cache_dir)
        return decorator(brain)

    # noinspection DuplicatedCode
    def decorator(func):
        func = check_if_function_is_method(func)
        cache_func = read_cache(func, cache_dir)
        if cache and cache_func.get('src', None) == get_source(func):
            func.__doc__ = json.dumps(cache_func.get('doc'), ensure_ascii=False)
            return func
        return docstring_generator(func, brain, cache, cache_dir)
    return decorator



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("LICENSE", "r", encoding="utf-8") as fh:
    license_content = fh.read()

setuptools.setup(
    name="ceo-py",
    version='0.13.0-preview',
    author="vortezwohl",
    author_email="vortez.wohl@gmail.com",
    description="An ultra-lightweight agentic AI framework based on the ReAct paradigm.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=license_content,
    url="https://github.com/vortezwohl/CEO",
    project_urls={
        "Bug Tracker": "https://github.com/vortezwohl/CEO/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    python_requires=">=3.10",
    install_requires=[
        'langchain>=0.3.14',
        'langchain-core>=0.3.29',
        'langchain-community>=0.3.14',
        'langchain-openai>=0.2.3',
        'openai>=1.59.3',
        'dashscope>=1.20.14',
        'mcp>=1.6.0'
    ],
    entry_points={},
    include_package_data=False
)



import json
import logging
from collections import OrderedDict

from typing_extensions import override

from ceo.ability import Ability
from ceo.brain.memory_augment import MemoryAugment

PREFIX = '__AgenticAbility__'

from ceo.brain.hook.after_action_taken import BaseHook
from ceo.brain.base_agent import BaseAgent

log = logging.getLogger('ceo.ability')


class AgenticAbility(Ability):
    def __init__(self, agent: BaseAgent | MemoryAugment):
        try:
            _test_mem = agent.memory
        except AttributeError:
            raise TypeError("The 'agent' of AgenticAbility should be instance of 'ceo.Agent'.")
        self._agent = agent
        self.__name__ = f'{PREFIX}talk_to_{agent.name}'
        self.__doc__ = json.dumps({
            "description": {
                "brief_description": f'Initiates a conversation with "{agent.name}" to use its abilities.',
                "detailed_description": f"First, carefully consider and explore {agent.name}'s "
                                        "potential abilities in solving your tasks, "
                                        f"then, if you need {agent.name}'s help, "
                                        "you must tell comprehensively, precisely "
                                        f"and exactly what you need {agent.name} to do.",
                f"self_introduction_from_{agent.name}": agent.introduction,
                "hint": f"By reading <self_introduction_from_{agent.name}>, you can learn what {agent.name} can do, "
                        f"and then decide whether to initiates a conversation with {agent.name} "
                        "according to its abilities.",
                "parameters": [],
                # "parameters": [{
                #     "request": {
                #         "name": "request",
                #         "type": "str",
                #         "description": f"A brief instruction to be processed by {agent.name}."
                #     },
                #     "request_by_step": {
                #         "name": "request_by_step",
                #         "type": "str",
                #         "description": "A comprehensive, precise and exact instruction "
                #                        f"to be processed by {agent.name}."
                #     },
                #     "memory": {
                #         "name": "memory",
                #         "type": "dict",
                #         "description": "/"
                #     }
                # }],
                "returns": {
                    "type": "str",
                    "description": f"{agent.name}'s response to your instruction."
                }
            }
        }, ensure_ascii=False)
        super().__init__(self)
        log.debug(f'Agent dispatcher generated. {self.__name__}: {self.__doc__}')

    @override
    def __call__(self, *args, **kwargs) -> str:
        _do_nothing = 'Do nothing.'
        request = kwargs.get('request', _do_nothing)
        request_by_step = kwargs.get('request_by_step', _do_nothing)
        memory = kwargs.get('memory', OrderedDict())
        before_action_taken_hook = kwargs.get('before_action_taken_hook', None)
        after_action_taken_hook = kwargs.get('after_action_taken_hook', None)
        if before_action_taken_hook is None:
            before_action_taken_hook = BaseHook.do_nothing()
        if after_action_taken_hook is None:
            after_action_taken_hook = BaseHook.do_nothing()
        self._agent.relay(request_by_step=request_by_step, request=request)
        self._agent.bring_in_memory(memory)
        return self._agent.just_do_it(before_action_taken_hook, after_action_taken_hook).response_for_agent



import logging

from ceo.brain.agent import Agent
from ceo.ability.agentic_ability import AgenticAbility

log = logging.getLogger('ceo.ability')


def agentic(agent: Agent):
    return lambda func: AgenticAbility(agent)



from typing_extensions import override

from mcp import ClientSession
from mcp.types import Tool, CallToolResult

from ceo.ability.base_ability import BaseAbility

DEFAULT_RETURNS = "<class 'str'>"
NONE = '/'


class McpAbility(BaseAbility):
    def __init__(self, mcp_tool: Tool, session: ClientSession):
        self._session: ClientSession = session
        _descr = {
            'brief_description': mcp_tool.description,
            'parameters': [
                {'parameter_name': k, 'type': v.get("type", DEFAULT_RETURNS),
                 'description': v.get('description', NONE)}
                for k, v in mcp_tool.inputSchema.get('properties', dict()).items()
            ]
        }
        super().__init__(
            name=mcp_tool.name,
            description=_descr,
            parameters=mcp_tool.inputSchema.get('properties', dict()),
            returns=DEFAULT_RETURNS
        )

    async def _call_tool(self, *args, **kwargs) -> CallToolResult:
        return await self.session.call_tool(name=self.name, arguments=kwargs)

    @override
    async def __call__(self, *args, **kwargs):
        _res = await self._call_tool(*args, **kwargs)
        _status = 'SUCCESSFUL' if not _res.isError else 'FAILED'
        return (f'McpAbility(mcp_tool="{self.name}") was {_status}.\n'
                f'Result: "{_res.content}"')

    @property
    def session(self) -> ClientSession:
        return self._session
